# Copyright 2021 The LeafNLP Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import pickle
import re
import shutil
import time
from glob import glob
from pprint import pprint

import numpy as np
import torch
from leafnlp.LeafBase.utils.data_utils import create_batch_file
from leafnlp.LeafBase.utils.utils import show_progress


class End2EndBase(object):
    """
    This engine is for the end2end training for seq2seq models.
    It is a new version of previous one. 
    Our goal is to extend its application to all kinds of language generation tasks.
    """

    def __init__(self, args=None):
        """
        Initialize
        """
        self.args = args
        self.base_models = {}
        self.train_models = {}
        self.pretrained_models = {}
        self.batch_data = {}
        self.global_steps = 0

    def build_vocabulary(self):
        """
        vocabulary
        """
        raise NotImplementedError

    def build_scheduler(self, optimizer):
        """
        Schedule Learning Rate
        """
        scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer=optimizer, step_size=self.args.step_size,
            gamma=self.args.step_decay)

        return scheduler

    def build_models(self):
        """
        Models:
            self.base_models: models that will be trained
                Format: {'name1': model1, 'name2': model2}
            self.train_models: models that will be trained.
                Format: {'name1': model1, 'name2': model2}
        """
        raise NotImplementedError

    def init_base_model_params(self):
        """
        Initialize Base Model Parameters.
        """
        for model_name in self.base_models:
            fl_ = "{}/{}.model".format(
                self.args.base_model_dir, model_name)
            self.base_models[model_name].load_state_dict(
                torch.load(fl_, map_location=lambda storage, loc: storage))

    def init_train_model_params(self):
        """
        Initialize Train Model Parameters.
        For testing and visulization.
        """
        for model_name in self.train_models:
            fl_ = "{}/{}.model".format(
                self.args.train_model_dir, model_name)
            self.train_models[model_name].load_state_dict(
                torch.load(fl_, map_location=lambda storage, loc: storage))

    def build_pipelines(self):
        """
        Pipelines and loss here.
        """
        raise NotImplementedError

    def build_optimizer(self, params):
        """
        define optimizer
        """
        raise NotImplementedError

    def load_batch(self, file_):
        """
        process batch data.
        """
        data = []
        fp = open(file_, "r")
        for line in fp:
            data.append(line)
        fp.close()

        return data

    def build_batch(self, batch_):
        """
        process batch data.
        """
        raise NotImplementedError

    def test_worker(self):
        """
        Used in decoding.
        Users can define their own decoding process.
        You do not have to worry about path and prepare input.
        """
        raise NotImplementedError

    def app_worker(self):
        """
        For application.
        """
        raise NotImplementedError

    def train(self):
        """
        training here.
        Don't overwrite.
        """
        out_dir = self.args.work_dir
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)
        if not os.path.exists("{}/model".format(out_dir)):
            os.mkdir("{}/model".format(out_dir))

        self.build_vocabulary()
        self.build_models()
        pprint(self.base_models)
        pprint(self.train_models)
        self.init_base_model_params()
        # here it is necessary to put list. Instead of directly append.
        params = []
        for model_name in self.train_models:
            params += list(self.train_models[model_name].parameters())
        if self.args.train_base_model:
            for model_name in self.base_models:
                params += list(self.base_models[model_name].parameters())
        print('Total number of parameters (Not Accurate): {}.'.format(
            sum([para.numel() for para in params])))
        # define optimizer
        optimizer = self.build_optimizer(params)
        if self.args.lr_schedule == 'decay':
            scheduler = self.build_scheduler(optimizer)
        # load checkpoint
        uf_model = [0, -1]
        if self.args.continue_training:
            model_para_files = glob(os.path.join(out_dir, '*.model'))
            if len(model_para_files) > 0:
                uf_model = []
                for fl_ in model_para_files:
                    arr = re.split(r'\/', fl_)[-1]
                    arr = re.split(r'\_|\.', arr)
                    arr = [int(arr[-3]), int(arr[-2])]
                    if arr not in uf_model:
                        uf_model.append(arr)
                cc_model = sorted(uf_model)[-1]
                try:
                    print("Try *_{}_{}.model".format(cc_model[0], cc_model[1]))
                    for model_name in self.train_models:
                        fl_ = "{}/{}_{}_{}.model".format(
                            out_dir, model_name, cc_model[0], cc_model[1])
                        self.train_models[model_name].load_state_dict(
                            torch.load(fl_, map_location=lambda storage, loc: storage))
                except:
                    cc_model = sorted(uf_model)[-2]
                    print("Try *_{}_{}.model".format(cc_model[0], cc_model[1]))
                    for model_name in self.train_models:
                        fl_ = "{}/{}_{}_{}.model".format(
                            out_dir, model_name, cc_model[0], cc_model[1])
                        self.train_models[model_name].load_state_dict(
                            torch.load(fl_, map_location=lambda storage, loc: storage))
                print('Continue training with *_{}_{}.model'.format(
                    cc_model[0], cc_model[1]))
                uf_model = cc_model
        else:
            shutil.rmtree(out_dir)
            os.mkdir(out_dir)
            os.mkdir("{}/model".format(out_dir))
        # train models
        fout = open('{}/args.pickled'.format(self.args.work_dir), 'wb')
        pickle.dump(self.args, fout)
        fout.close()
        start_time = time.time()
        cclb = 0
        for epoch in range(uf_model[0], self.args.n_epoch):
            n_batch = create_batch_file(
                path_data=self.args.data_dir,
                path_work=self.args.work_dir,
                is_shuffle=True,
                fkey_=self.args.task,
                file_=self.args.file_train,
                batch_size=self.args.batch_size,
                is_lower=self.args.is_lower)
            print('The number of batches: {}'.format(n_batch))
            self.global_steps = n_batch * max(0, epoch)
            for batch_id in range(n_batch):
                self.global_steps += 1
                learning_rate = self.args.learning_rate
                if self.args.lr_schedule == 'warm-up':
                    learning_rate = 2.0 * \
                        (self.args.model_size ** (-0.5) *
                         min(self.global_steps ** (-0.5),
                             self.global_steps * self.args.warmup_step**(-1.5)))
                    for p in optimizer.param_groups:
                        p['lr'] = learning_rate
                elif self.args.lr_schedule == 'decay':
                    for p in optimizer.param_groups:
                        learning_rate = p['lr']
                        break
                if cclb == 0 and batch_id < n_batch-1 and batch_id <= uf_model[1]:
                    continue
                else:
                    cclb += 1

                fl_ = "{}/batch_{}_{}/{}".format(
                    self.args.work_dir, self.args.task,
                    self.args.batch_size, batch_id)
                self.build_batch(self.load_batch(fl_))
                loss = self.build_pipelines()

                optimizer.zero_grad()
                loss.backward()
                torch.nn.utils.clip_grad_norm_(params, self.args.grad_clip)
                optimizer.step()

                end_time = time.time()
                if batch_id % self.args.checkpoint == 0:
                    for model_name in self.train_models:
                        if epoch >= 0.8*self.args.n_epoch \
                                or epoch == self.args.n_epoch-1 \
                                or epoch == self.args.n_epoch-2:
                            fl_ = "{}/{}_{}_{}.model".format(
                                out_dir, model_name, epoch, batch_id)
                            fmodel = open(fl_, "wb")
                        else:
                            fl_ = "{}/{}_{}_0.model".format(
                                out_dir, model_name, epoch)
                            fmodel = open(fl_, 'wb')
                        torch.save(
                            self.train_models[model_name].state_dict(), fmodel)
                        fmodel.close()
                        fl_ = "{}/model/{}.model".format(out_dir, model_name)
                        fmodel = open(fl_, "wb")
                        torch.save(
                            self.train_models[model_name].state_dict(), fmodel)
                        fmodel.close()
                if batch_id % 1 == 0:
                    end_time = time.time()
                    print('epoch={}, batch={}, lr={}, loss={}, time={}h'.format(
                        epoch, batch_id, np.around(learning_rate, 6),
                        np.round(float(loss.data.cpu().numpy()), 4),
                        np.round((end_time-start_time)/3600.0, 4)))
                del loss
            for model_name in self.train_models:
                if epoch >= 0.8*self.args.n_epoch \
                        or epoch == self.args.n_epoch-1 \
                        or epoch == self.args.n_epoch-2:
                    fl_ = "{}/{}_{}_{}.model".format(
                        out_dir, model_name, epoch, batch_id)
                    fmodel = open(fl_, 'wb')
                else:
                    fl_ = "{}/{}_{}_0.model".format(
                        out_dir, model_name, epoch)
                    fmodel = open(fl_, 'wb')
                torch.save(
                    self.train_models[model_name].state_dict(), fmodel)
                fmodel.close()
                fl_ = "{}/model/{}.model".format(out_dir, model_name)
                fmodel = open(fl_, "wb")
                torch.save(
                    self.train_models[model_name].state_dict(), fmodel)
                fmodel.close()
            if self.args.lr_schedule == 'decay':
                scheduler.step()

        files_ = glob('{}/batch_train_*'.format(self.args.work_dir))
        for fl in files_:
            shutil.rmtree(fl)

    def validate(self):
        """
        Validation here.
        Don't overwrite.
        """
        self.build_vocabulary()
        self.build_models()
        pprint(self.base_models)
        pprint(self.train_models)
        self.init_base_model_params()

        best_arr = []
        val_file = os.path.join(self.args.work_dir, "model_validate.txt")
        if os.path.exists(val_file):
            fp = open(val_file, 'r')
            for line in fp:
                arr = re.split(r'\s', line[:-1])
                best_arr.append(
                    [arr[0], arr[1], arr[2], float(arr[3]), float(arr[4])])
            fp.close()

        for model_name in self.base_models:
            self.base_models[model_name].eval()
        for model_name in self.train_models:
            self.train_models[model_name].eval()
        with torch.no_grad():
            while 1:
                model_para_files = []
                model_para_files = glob(os.path.join(
                    self.args.work_dir, 
                    sorted(list(self.train_models))[0]+'*.model'))
                for j in range(len(model_para_files)):
                    arr = re.split(r'\_|\.', model_para_files[j])
                    arr = [int(arr[-3]), int(arr[-2]), model_para_files[j]]
                    model_para_files[j] = arr
                model_para_files = sorted(model_para_files)

                for fl_ in model_para_files:
                    best_model = {itm[0]: itm[3] for itm in best_arr}
                    if fl_[-1] in best_model:
                        continue
                    print('Validate *_{}_{}.model'.format(fl_[0], fl_[1]))

                    losses = []
                    start_time = time.time()
                    if os.path.exists(fl_[-1]):
                        time.sleep(3)
                        try:
                            for model_name in self.train_models:
                                fl_tmp = "{}/{}_{}_{}.model".format(
                                    self.args.work_dir, model_name, fl_[0], fl_[1])
                                self.train_models[model_name].load_state_dict(
                                    torch.load(fl_tmp, map_location=lambda storage, loc: storage))
                        except:
                            print('Models cannot be load!!!')
                            continue
                    else:
                        continue
                    val_batch = create_batch_file(
                        path_data=self.args.data_dir,
                        path_work=self.args.work_dir,
                        is_shuffle=True,
                        fkey_=self.args.task,
                        file_=self.args.file_val,
                        batch_size=self.args.batch_size, 
                        is_lower=self.args.is_lower)
                    print('The number of batches (Dev): {}'.format(val_batch))
                    if self.args.val_n_batches > val_batch:
                        self.args.val_n_batches = val_batch
                    for batch_id in range(self.args.val_n_batches):
                        
                        fl2_ = "{}/batch_{}_{}/{}".format(
                            self.args.work_dir, self.args.task,
                            self.args.batch_size, batch_id)
                        self.build_batch(self.load_batch(fl2_))
                        loss = self.build_pipelines()

                        losses.append(loss.data.cpu().numpy())
                        show_progress(batch_id+1, self.args.val_n_batches)
                    print()
                    losses = np.array(losses)
                    end_time = time.time()
                    if self.args.use_move_avg:
                        try:
                            losses_out = 0.9*losses_out + \
                                0.1*np.average(losses)
                        except:
                            losses_out = np.average(losses)
                    else:
                        losses_out = np.average(losses)
                    best_arr.append(
                        [fl_[2], fl_[0], fl_[1], losses_out, end_time-start_time])
                    best_arr = sorted(best_arr, key=lambda bb: bb[3])
                    if best_arr[0][0] == fl_[2]:
                        out_dir = os.path.join(self.args.work_dir, 'model')
                        for model_name in self.base_models:
                            fmodel = open(os.path.join(
                                out_dir, model_name+'.model'), 'wb')
                            torch.save(
                                self.base_models[model_name].state_dict(), fmodel)
                            fmodel.close()
                        for model_name in self.train_models:
                            fmodel = open(os.path.join(
                                out_dir, model_name+'.model'), 'wb')
                            torch.save(
                                self.train_models[model_name].state_dict(), fmodel)
                            fmodel.close()
                        try:
                            shutil.copy2(os.path.join(
                                self.args.data_dir, self.args.file_vocab), out_dir)
                        except:
                            pass
                    for itm in best_arr[:self.args.n_best_models]:
                        print('model={}_{}, loss={}, time={}'.format(
                            itm[1], itm[2], np.round(float(itm[3]), 4),
                            np.round(float(itm[4]), 4)))

                    for itm in best_arr[self.args.n_best_models:]:
                        tarr = re.split(r'_|\.', itm[0])
                        if tarr[-2] == '0':
                            continue
                        if os.path.exists(itm[0]):
                            for model_name in self.train_models:
                                fl_tmp = "{}/{}_{}_{}.model".format(
                                    self.args.work_dir, model_name, itm[1], itm[2])
                                os.unlink(fl_tmp)
                    fout = open(val_file, "w")
                    for itm in best_arr:
                        if len(itm) == 0:
                            continue
                        fout.write(' '.join([itm[0], str(itm[1]), str(
                            itm[2]), str(itm[3]), str(itm[4])])+'\n')
                    fout.close()
                if self.args.validate_once:
                    break

        files_ = glob('{}/batch_validate_*'.format(self.args.work_dir))
        for fl in files_:
            shutil.rmtree(fl)

    def test(self):
        """
        testing
        Don't overwrite.
        """
        self.build_vocabulary()
        self.build_models()
        pprint(self.base_models)
        pprint(self.train_models)
        self.init_base_model_params()

        _nbatch = create_batch_file(
            path_data=self.args.data_dir,
            path_work=self.args.work_dir,
            is_shuffle=False,
            fkey_=self.args.task,
            file_=self.args.file_test,
            batch_size=self.args.batch_size, 
            is_lower=self.args.is_lower)
        print('The number of samples (test): {}'.format(_nbatch))

        for model_name in self.base_models:
            self.base_models[model_name].eval()
        for model_name in self.train_models:
            self.train_models[model_name].eval()
        with torch.no_grad():
            for model_name in self.train_models:
                if self.args.use_optimal_model:
                    model_file = "{}/model/{}.model".format(
                        self.args.work_dir, model_name)
                else:
                    arr = re.split(r'\D', self.args.model_optimal_key)
                    model_file = "{}/{}_{}_{}.model".format(
                        self.args.work_dir, model_name, arr[1], arr[2])
                self.train_models[model_name].load_state_dict(torch.load(
                    model_file, map_location=lambda storage, loc: storage))

            start_time = time.time()
            output_file = os.path.join(
                self.args.work_dir, self.args.file_output)

            fout = open(output_file, 'w')
            for batch_id in range(_nbatch):
                fl2_ = "{}/batch_{}_{}/{}".format(
                    self.args.work_dir, self.args.task,
                    self.args.batch_size, batch_id)
                self.build_batch(self.load_batch(fl2_))
                test_data = self.test_worker()
                for data2dump in test_data:
                    json.dump(data2dump, fout)
                    fout.write('\n')

                end_time = time.time()
                show_progress(batch_id, _nbatch, str(
                    (end_time-start_time)/3600)[:8]+"h")
            fout.close()
            print()

        files_ = glob('{}/batch_test_*'.format(self.args.work_dir))
        for fl in files_:
            shutil.rmtree(fl)
