# Copyright 2021 The LeafNLP Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import torch


class LayerNormalization(torch.nn.Module):
    """
    Transformer Layer Normalization.
    """

    def __init__(self, size, eps=1e-5, bias=False):
        super().__init__()

        self.eps = eps
        self.bias = bias

        self.gamma = torch.nn.Parameter(torch.ones(size))
        self.register_parameter('gamma', self.gamma)

        if bias:
            self.beta = torch.nn.Parameter(torch.zeros(size))
            self.register_parameter('beta', self.beta)

    def forward(self, input_):
        """
        gamma * (input - mean) / (std + eps) + bias
        """
        mean = torch.mean(input_, -1, keepdim=True)
        std = torch.mean((input_-mean)*(input_-mean), -1, keepdim=True)
        
        output = self.gamma*(input_-mean)/torch.sqrt(std + self.eps)
        if self.bias:
            output = output + self.beta

        return output
