# Copyright 2021 The LeafNLP Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import torch
from torch.autograd import Variable

from ..seqlabel_base import PipeSeqLabelBase
from .model import modelNERBERT


class trainerSeqLabel(modelNERBERT, PipeSeqLabelBase):

    def __init__(self, args):
        self.args = args
        PipeSeqLabelBase.__init__(self, args=args)
        modelNERBERT.__init__(self, 
                               task=args.task,
                               base_model_dir=args.base_model_dir, 
                               train_model_dir=args.train_model_dir)
        
        self.loss_criterion = torch.nn.CrossEntropyLoss(
            ignore_index=-1).to(self.device)
    
    def build_vocabulary(self):
        modelNERBERT.build_vocabulary(self)
        
        fl1_ = "{}/model/label_mapping.json".format(self.args.work_dir)
        fl2_ = "{}/label_mapping.json".format(self.args.data_dir)
        with open(fl2_, "r") as fp:
            label_mapping = json.load(fp)
        with open(fl1_, "w") as fout:
            json.dump(label_mapping, fout)
        self.label2id = {key: label_mapping[key] for key in label_mapping}
        self.id2label = {label_mapping[key]: key for key in label_mapping}
        self.n_labels = len(self.label2id)
        
    def build_batch(self, batch_):
        """
        get batch data
        """
        modelNERBERT.build_batch(self, batch_)
        
        label_arr = []
        for line in batch_:
            itm = json.loads(line)
            if len(itm["tokens"]) > 300:
                continue
            tmp_label = []
            for lb in itm["labels"]:
                tmp_label.append(self.label2id[lb])
            tmp_label = [-1] + tmp_label + [-1]
            label_arr.append(tmp_label)

        maxlen_text = self.batch_data["maxlen_text"]
        label_arr = [itm[:maxlen_text] for itm in label_arr]
        label_arr = [itm + [-1 for _ in range(maxlen_text-len(itm))]
                        for itm in label_arr]
        label_var = Variable(torch.LongTensor(label_arr))

        self.batch_data['labels'] = label_var.to(self.device)
    
    def build_pipelines(self):
        """
        Data flow from input to output.
        """
        logits = self.build_pipe()
        logits = logits.contiguous().view(-1, self.n_labels)

        loss = self.loss_criterion(
            logits, self.batch_data['labels'].view(-1))

        return loss
    
    def test_worker(self):
        """
        Testing.
        """
        logits = self.build_pipe()
        logits = torch.softmax(logits, dim=2)

        pred = logits.topk(1, dim=2)[1].squeeze(2)
        pred = pred.data.cpu().numpy().tolist()
        for k, itm in enumerate(self.batch_data["output"]):

            label = [self.id2label[lb] for lb in pred[k]]
            label = label[1:]
            label = label[:len(itm["tokens"])]
            itm["labels_pred"] = label

            yield itm
